# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import sys
import os
import pdb
from numpy import inf

from QPalma.sequence_utils import SeqSpliceInfo,get_flat_file_size,DataAccessWrapper,reverse_complement


class LookupTable:
   """
   Speed up genomic and splice site score queries by prefetching the sequences
   and the scores once for all chromosomes.
   """

   def __init__(self,settings):
      chromo_list = settings['allowed_fragments']
      accessWrapper = DataAccessWrapper(settings)
      self.seqInfo = SeqSpliceInfo(accessWrapper,settings['allowed_fragments'])

      # take care that it also works say if we get a chromo_list [1,3,5]
      # those are mapped then to 0,1,2
      self.chromo_list = chromo_list
      self.chromo_map = {}
      for idx,elem in enumerate(chromo_list):
         self.chromo_map[elem] = idx

      #print 'Chromo mapping is %s' % str(self.chromo_map)

      self.genomicSequences   = {}
      self.acceptorScores     = {}
      self.donorScores        = {}

      self.genomicSequences = [0]*(len(self.chromo_list))
      self.acceptorScoresPos   = [0]*(len(self.chromo_list))
      self.donorScoresPos      = [0]*(len(self.chromo_list))
      self.acceptorScoresNeg   = [0]*(len(self.chromo_list))
      self.donorScoresNeg      = [0]*(len(self.chromo_list))

      for chrId in self.chromo_list:
         self.prefetch_seq_and_scores(chrId, settings)

 
   def get_seq_and_scores(self,chromo,strand,start,stop):
      assert chromo in self.chromo_list
      assert start <= stop

      chromo_idx = self.chromo_map[chromo]


      if strand == '+':
         currentSequence         = self.genomicSequences[chromo_idx][start:stop]
         currentAcceptorScores   = self.acceptorScoresPos[chromo_idx][start:stop]
         currentDonorScores      = self.donorScoresPos[chromo_idx][start:stop]
      elif strand == '-':
         currentSequence         = self.genomicSequences[chromo_idx][start:stop]
         currentSequence = reverse_complement(currentSequence)

         total_size = self.seqInfo.getFragmentSize(chromo)
         _start = total_size - stop
         _stop = total_size - start
         currentAcceptorScores   = self.acceptorScoresNeg[chromo_idx][_start:_stop]
         currentDonorScores      = self.donorScoresNeg[chromo_idx][_start:_stop]

         #gt_tuple_pos = [p for p,e in enumerate(currentSequence) if (p>0 and p<len(currentSequence)-1 and e=='g' ) and (currentSequence[p+1]=='t' or currentSequence[p+1]=='c')]
         #ag_tuple_pos = [p for p,e in enumerate(currentSequence) if p>1 and currentSequence[p-1]=='a' and currentSequence[p]=='g']
         #acc_pos = [p for p,e in enumerate(currentAcceptorScores) if e != -inf and p > 1]
         #don_pos = [p for p,e in enumerate(currentDonorScores) if e != -inf and p > 0]

      return currentSequence, currentAcceptorScores, currentDonorScores


   def prefetch_seq_and_scores(self,chromo,settings):
      """
      This function expects an interval, chromosome and strand information and
      returns then the genomic sequence of this interval and the associated scores.
      """
      assert chromo in self.chromo_list

      genomicSeq_start  = 0
      genomicSeq_stop   = self.seqInfo.getFragmentSize(chromo)

      #print 'lt total_size %d' % self.seqInfo.getFragmentSize(chromo)

      if settings['pos_strand_prefetching_enabled']:
         genomicSeq,currentAcc,currentDon = self.seqInfo.get_seq_and_scores(chromo, '+', genomicSeq_start, genomicSeq_stop, False, settings['perform_checks'])
      else:
         genomicSeq,currentAcc,currentDon = None,None,None

      genomicSeq = genomicSeq.lower()

      chromo_idx = self.chromo_map[chromo]

      self.genomicSequences[chromo_idx] = genomicSeq
      self.acceptorScoresPos[chromo_idx]   = currentAcc
      self.donorScoresPos[chromo_idx]      = currentDon

      if settings['neg_strand_prefetching_enabled']:
         print 'prefetching neg strand info'
         dummy,currentAcc,currentDon = self.seqInfo.get_seq_and_scores(chromo, '-', genomicSeq_start, genomicSeq_stop, False, settings['perform_checks'])
      else:
         dummy,currentAcc,currentDon = None,None,None

      #genomicSeq = reverse_complement(self.genomicSequences[chromo_idx])
      #gt_tuple_pos = [p for p,e in enumerate(genomicSeq) if (p>0 and p<len(genomicSeq)-1 and e=='g' ) and (genomicSeq[p+1]=='t' or genomicSeq[p+1]=='c')]
      #ag_tuple_pos = [p for p,e in enumerate(genomicSeq) if p>1 and genomicSeq[p-1]=='a' and genomicSeq[p]=='g']
      #acc_pos = [p for p,e in enumerate(currentAcc) if e != -inf and p > 1]
      #don_pos = [p for p,e in enumerate(currentDon) if e != -inf and p > 0]
      #pdb.set_trace()

      self.acceptorScoresNeg[chromo_idx]   = currentAcc
      self.donorScoresNeg[chromo_idx]      = currentDon
