# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import cPickle
import math
import numpy
import os
import os.path
import pdb
import sys

from QPalma.utils import pprint_alignment
from QPalma.sequence_utils import SeqSpliceInfo,DataAccessWrapper


# how aligned est and dna sequence is displayed (gap_char, 5 nucleotides, intron_char) 
translation = '-acgtn_' 

# convenience function for pretty-printing arrays
pp = lambda x : str(x)[1:-1].replace(' ','')


def alignment_reconstruct(current_prediction,num_exons):
   """
   We reconstruct the exact alignment to infer the positions of indels and the
   sizes of the respective exons.
   """

   SpliceAlign = current_prediction['spliceAlign']
   estAlign    = current_prediction['estAlign']

   dna_array   = current_prediction['dna_array']
   read_array  = current_prediction['read_array']

   dna_array   = "".join(map(lambda x: translation[int(x)],dna_array))
   read_array  = "".join(map(lambda x: translation[int(x)],read_array))

   # this array holds a number for each exon indicating its number of matches
   exonIdentities = [0]*num_exons
   exonGaps       = [0]*num_exons

   gaps     = 0
   identity = 0
   idx      = 0

   for ridx in range(len(read_array)):
      read_elem = read_array[ridx]
      dna_elem = dna_array[ridx]

      if ridx > 0 and read_array[ridx-1] != '_' and read_array[ridx] == '_':
         if ridx > 1 and read_array[ridx-2] == '_' and read_array[ridx-1] != '_' and read_array[ridx] == '_':
            continue

         idx += 1
         gaps = 0
         identity = 0
         
      if read_elem == '_':
         continue

      if read_elem == dna_elem:
         identity += 1

      if read_elem == '-':
         gaps += 1

      if idx > num_exons-1: 
         pdb.set_trace()
      else:
         exonIdentities[idx] = identity
         exonGaps[idx] = gaps

   return exonIdentities,exonGaps


def getUniquePrediction(allPredictions):
   """
   Since one read can have several alignments due to several possible seed
   positions we have to preprocess all predictions. This function returns for
   every read only one alignment that is the highest scoring under the QPalma
   model.
   """

   allUniquePredictions = {}

   for new_prediction in allPredictions:
      id = new_prediction['id']
      if allUniquePredictions.has_key(id):
         current_prediction = allUniquePredictions[id]

         current_a_score = current_prediction['DPScores'].flatten().tolist()[0][0]
         new_score =  new_prediction['DPScores'].flatten().tolist()[0][0]

         if current_a_score < new_score :
            allUniquePredictions[id] = new_prediction

      else:
         allUniquePredictions[id] = new_prediction

   return allUniquePredictions 


def recalculatePositions(chromo,start_pos,strand,starts,ends,seqInfo,window_size):
   """
   If we work on the negative strand we have to recalculate the indices other
   wise we just have to add an offset i.e. the start position.
   """

   #print 'window_size is %d' % window_size
   #print start_pos
   #print starts
   #print ends

   if strand == '-':
      _ends = [window_size - int(elem) + 2 for elem in starts]
      _starts  = [window_size - int(elem) + 2 for elem in ends]
      starts = _starts
      ends = _ends

   #print start_pos
   #print starts
   #print ends

   starts   = [int(elem) + start_pos-1 for elem in starts]
   ends     = [int(elem) + start_pos-1 for elem in ends]

   return starts,ends


def createAlignmentOutput(settings,chunk_fn,result_fn):
   """
   Given all the predictions and a result file name. This function calculates
   all aligmnment information and returns the result in the specified format.
   """

   out_fh = open(result_fn,'w+')
   allPredictions = cPickle.load(open(chunk_fn))

   # fetch the data needed
   accessWrapper = DataAccessWrapper(settings)
   seqInfo = SeqSpliceInfo(accessWrapper,settings['allowed_fragments'])

   # make predictions unique -> take the highest scoring alignment for a given read
   allUniquePredictions = getUniquePrediction(allPredictions)

   output_format  = settings['output_format']
   
   if output_format == 'QPalma':
      # blat output
      new_line = createQPalmaBEDOutput(allUniquePredictions,settings,out_fh)

   if output_format == 'blat':
      # blat output
      new_line = createQPalmaBEDOutput(allUniquePredictions,settings,out_fh)
      #new_line = createBlatOutput(allUniquePredictions,settings,out_fh)

   elif output_format == 'ShoRe':
      # ShoRe output
      new_line = createShoReOutput(allUniquePredictions,settings)

   elif output_format == 'mGene':
      # Genefinding output for mGene
      new_line = createGenefindingOutput(allUniquePredictions,settings)

   else:
      assert False

   out_fh.close()


def createQPalmaBEDOutput(all_predictions,settings,out_fh):
   """
   This function takes one prediction as input and returns the corresponding
   alignment in blat format:


   """

   print 'Creating QPALMA/bed output...'

   # fetch the data needed
   allowed_fragments = settings['allowed_fragments']
   accessWrapper     = DataAccessWrapper(settings)
   seqInfo           = SeqSpliceInfo(accessWrapper,allowed_fragments)

   # this is the total size of the seed region
   window_size = 2*settings['half_window_size']

   written_lines = 0

   for pred_idx,current_prediction in enumerate(all_predictions.values()):
      # now we fetch the data of the read itself
      id = current_prediction['id']

      seq         = current_prediction['read']
      quality     = current_prediction['quality']
      dna         = current_prediction['dna']

      chromo      = current_prediction['chr']
      strand      = current_prediction['strand']
      start_pos   = current_prediction['start_pos']
      alignment   = current_prediction['alignment']

      DPScores    = current_prediction['DPScores']

      (qStart, qEnd, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds, tExonSizes,tStarts, tEnds) = alignment

      if len(qExonSizes) != num_exons:
         print 'BUG exon sizes %s'%id

      exonIdentities,exonGaps = alignment_reconstruct(current_prediction,num_exons)

      tStarts,tEnds = recalculatePositions(chromo,start_pos,strand,tStarts,tEnds,seqInfo,window_size)

      numMatches=len(seq)
      minExonLen=len(seq)
      for i in range(len(tExonSizes)):
          if minExonLen<tExonSizes[i]: minExonLen=tExonSizes[i]
      maxIntronLen=0 
      for i in range(len(tExonSizes)-1):
          if maxIntronLen>tStarts[i+1]-tEnds[i]: maxIntronLen=tStarts[i+1]-tEnds[i]
      readOrientation='+'
      numGaps=0
      
      # 1. contig name
      # 2. start position
      # 3. end position
      # 4. read_id
      # 5. number of matches
      # 6. strand
      # 7. number of unspliced alignment
      # 8. number of splice alignments
      # 9. 0,0,0 (not used; see bed format spec)
      # 10. number of exons
      # 11. block lengths
      # 12. relative block starts
      # 13. additional features
      #if id=='read_899': pdb.set_trace()
      
      new_line = '%s\t%d\t%d\t%s\t%d\t%s\t%d\t%d\t0,0,0\t%d\t%s\t%s\tqpalmaScore=%2.3f;numMatches=%d;numGaps=%d;minExonLen=%d;maxIntronLen=%d;readOrientation=%s;read=%s;quality=%s\n' %\
                 (chromo,tStarts[0],tEnds[-1],id,numMatches,strand,num_exons==1,num_exons>1,num_exons,pp(tExonSizes),pp([int(elem-tStarts[0]) for elem in tStarts]),DPScores[0],numMatches,numGaps,minExonLen,maxIntronLen,readOrientation,seq,quality)
      written_lines += 1

      out_fh.write(new_line)

   print 'Wrote out %d lines' % written_lines

def createBlatOutput(all_predictions,settings,out_fh):
   """
   This function takes one prediction as input and returns the corresponding
   alignment in blat format:

    1. matches     - Number of bases that match that aren't repeats
    2. misMatches  - Number of bases that don't match
    3. repMatches  - Number of bases that match but are part of repeats
    4. nCount      - Number of 'N' bases
    5. qNumInsert  - Number of inserts in query
    6. qBaseInsert - Number of bases inserted in query
    7. tNumInsert  - Number of inserts in target
    8. tBaseInsert - Number of bases inserted in target
    9. strand      - '+' or '-' for query strand. In mouse, second '+'or '-' is for genomic strand
   10. qName       - Query sequence name
   11. qSize       - Query sequence size
   12. qStart      - Alignment start position in query
   13. qEnd        - Alignment end position in query
   14. tName       - Target sequence name
   15. tSize       - Target sequence size
   16. tStart      - Alignment start position in target
   17. tEnd        - Alignment end position in target
   18. blockCount  - Number of blocks in the alignment
   19. blockSizes  - Comma-separated list of sizes of each block
   20. qStarts     - Comma-separated list of starting positions of each block in query
   21. tStarts     - Comma-separated list of starting positions of each block in target 

   """

   print 'Creating blat output...'

   # fetch the data needed
   allowed_fragments = settings['allowed_fragments']
   accessWrapper     = DataAccessWrapper(settings)
   seqInfo           = SeqSpliceInfo(accessWrapper,allowed_fragments)

   # this is the total size of the seed region
   window_size = 2*settings['half_window_size']

   written_lines = 0

   for pred_idx,current_prediction in enumerate(all_predictions.values()):
      # now we fetch the data of the read itself
      id = current_prediction['id']

      seq         = current_prediction['read']
      dna         = current_prediction['dna']

      chromo      = current_prediction['chr']
      strand      = current_prediction['strand']
      start_pos   = current_prediction['start_pos']
      alignment   = current_prediction['alignment']

      DPScores    = current_prediction['DPScores']

      (qStart, qEnd, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds,\
      tExonSizes,tStarts, tEnds) = alignment

      if len(qExonSizes) != num_exons:
         print 'BUG exon sizes %s'%id

      exonIdentities,exonGaps = alignment_reconstruct(current_prediction,num_exons)

      tStarts,tEnds = recalculatePositions(chromo,start_pos,strand,tStarts,tEnds,seqInfo,window_size)

      new_line = '%s\t%s\t%s\t%d\t%d\t%d\t%d\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' %\
      (id,chromo,strand,start_pos,qStart,qEnd,tStart,tEnd,num_exons,\
      pp(qExonSizes),pp(qStarts),pp(qEnds),pp(tExonSizes),\
      pp(tStarts),pp(tEnds),pp([int(elem) for elem in exonIdentities]),pp([int(elem) for elem in exonGaps]))
      written_lines += 1

      out_fh.write(new_line)

   print 'Wrote out %d lines' % written_lines


def createGenefindingOutput(current_loc,out_fname):
   """
   This is an output format used by mGene. It consists of positional informations to generate the exon and intron tracks.
   """

   #out_fh.write("%d\t%s\t%s\t%s\t%s\t%s\n"%(chromo,strand,pp(starts),pp(ends),pp(ids),pp(gaps)))
   pass


def createShoReOutput(current_loc,out_fname):
   """
   This format is used by the additional steps implemented in the ShoRe pipeline.
   """

   pass
