# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import sys
import os
import os.path
import logging
import math
import pdb

import cvxopt.base as cb

import logging
logging.basicConfig(level=logging.DEBUG,format='%(levelname)s %(message)s')

class SIQP:
   """
   This class is a wrapper for the cvxopt/cplex qp solvers.
   It has the purpose to support an iterative addition of
   constraints to the original problem.

   We want to solve a problem of the form

   min  1/2 * x' * P * x + C * q' * x
    x

    s.t. 
      < w,f+ > - < w,f- > >= 1 - xi_i    for all i, for all f-

   where x is a vector storing information on the parameters w_i and the slacks xi_i
   so x = [w;xi] 

   """

   def __init__(self,fSize,numExamples,settings):
      assert fSize > 0, 'You have to have at least one feature!'
      assert numExamples > 0, 'You have to have at least one example!'
      self.numFeatures = fSize
      self.numExamples = numExamples
      self.C = settings['C']
      self.EPSILON = 10e-15

      self.settings = settings

      logging.debug("Starting SIQP solver with %d examples. A feature space dim. of %d.", numExamples,fSize)
      logging.debug("Regularization parameters are C=%f."%(self.C))

      self.old_w = None

      self.dimP = self.numFeatures + self.numExamples
      self.createRegularizer()


   def enforceMonotonicityBetween(self,beg,end):
      G_rows,G_cols = self.G.size
      h_rows,h_cols = self.h.size

      add_rows = end - beg - 1
      if add_rows < 1:
         print 'Error: enforceMonotonicityBetween(beg,end) has to be called for at least an integer interval of size 1!'
         return 

      new_rows = G_rows + add_rows
      new_G = cb.matrix(0.0,(new_rows,G_cols))
      new_G[0:G_rows,:] = self.G[0:G_rows,:]

      new_h = cb.matrix(0.0,(new_rows,h_cols))
      new_h[0:h_rows,0] = self.h[0:h_rows,0]

      cur_row = 0
      for idx in range(beg,end-1):
         print cur_row
         new_G[G_rows+cur_row,idx]    =  1.0
         new_G[G_rows+cur_row,idx+1]  = -1.0
         cur_row += 1

      self.G = new_G
      self.h = new_h


   def createUnitRegularizer(self):
      # set regularizer to zero
      self.P = cb.matrix(0.0,(self.dimP,self.dimP))
      for i in range(self.numFeatures):
         self.P[i,i] = 1.0
      # end of zeroing regularizer


   def createSmoothnessRegularizer(self):
      settings = self.settings

      self.P = cb.matrix(0.0,(self.dimP,self.dimP))
      # for numerical stability
      for i in range(self.numFeatures):
         self.P[i,i] = 1e-5

      lengthSP    = settings['numLengthSuppPoints']
      donSP       = settings['numDonSuppPoints']
      accSP       = settings['numAccSuppPoints']
      mmatrixSP   = settings['matchmatrixRows']*settings['matchmatrixCols']
      numq        = settings['numQualSuppPoints']
      numQPlifs   = settings['numQualPlifs'] 

      intronLengthGroupParam  = 1.0
      donorGroupParam         = 1.0
      acceptorGroupParam      = 1.0
      matchmatrixGroupParam   = 1.0
      qualityGroupParam       = 1.0

      # add off-diagonal terms for intron length parameters
      for j in range(lengthSP-1):
         self.P[j,j+1]   += -intronLengthGroupParam
         self.P[j+1,j]   += -intronLengthGroupParam
         self.P[j,j]     +=  intronLengthGroupParam
         self.P[j+1,j+1] +=  intronLengthGroupParam

      # add off-diagonal terms for donor parameters
      for j in range(lengthSP,lengthSP+donSP-1):
         self.P[j,j+1]   += -donorGroupParam
         self.P[j+1,j]   += -donorGroupParam
         self.P[j,j]     +=  donorGroupParam
         self.P[j+1,j+1] +=  donorGroupParam
      
      # we enforce monotonic increasing donor scores
      self.enforceMonotonicityBetween(lengthSP,lengthSP+donSP)

      # add off-diagonal terms for acceptor parameters
      for j in range(lengthSP+donSP,lengthSP+donSP+accSP-1):
         self.P[j,j+1]   += -acceptorGroupParam
         self.P[j+1,j]   += -acceptorGroupParam
         self.P[j,j]     +=  acceptorGroupParam
         self.P[j+1,j+1] +=  acceptorGroupParam

      # we enforce monotonic increasing acceptor scores
      self.enforceMonotonicityBetween(lengthSP+donSP,lengthSP+donSP+accSP)

      # no off-diagonal terms for matchmatrix parameters
      for j in range(lengthSP+donSP+accSP,lengthSP+donSP+accSP+mmatrixSP):
          self.P[j,j] += matchmatrixGroupParam

      # add off-diagonal terms for quality parameters
      offset = lengthSP+donSP+accSP+mmatrixSP
      for k in range(numQPlifs):
         begin = offset+(k*numq)
         end   = offset+((k+1)*numq)

         for j in range(begin,end-1):
            self.P[j,j+1]   += -qualityGroupParam
            self.P[j+1,j]   += -qualityGroupParam
            self.P[j,j]     +=  qualityGroupParam
            self.P[j+1,j+1] +=  qualityGroupParam


   def createRegularizer(self):
      
      q_array = [0.0]*self.numFeatures + [1.0]*self.numExamples
      self.q = cb.matrix(q_array,(self.numFeatures+self.numExamples,1),'d')
      self.q = self.C * self.q

      # add constraints for non-negativity of slacks
      self.G = cb.matrix(0.0,(self.numExamples,self.numFeatures+self.numExamples))
      for i in range(self.numExamples):
         self.G[i,self.numFeatures+i] = -1

      self.h = cb.matrix(0,(self.numExamples,1),'d')

      # quadratic term creation should take place after above constraint system
      # generation because we also create monotonicity contraints in there
      self.createUnitRegularizer()
      #self.createSmoothnessRegularizer()



if __name__ == '__main__':
   settings = {}
   settings['numLengthSuppPoints'] = 5
   settings['numDonSuppPoints'] = 10
   settings['numAccSuppPoints'] = 10
   settings['matchmatrixRows'] = 1
   settings['matchmatrixCols'] = 6
   settings['numQualSuppPoints'] = 5
   settings['numQualPlifs'] = 30
   sq = SIQP(181,2,100.0,settings)

   """
   function [C_qp_penalties,C_qp_smoothness,C_lp_smoothness,C_qp_smallness,column_eps,Q,f,LB,UB] = paths_create_qp(N,C)
   % function [C_qp_penalties,C_qp_smoothness,C_lp_smoothness,C_qp_smallness,column_eps,Q,f,LB,UB] = paths_create_qp(N,C)

   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % qp_solve: min ( <res, f'> + 1/2 res' Q res)
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   C_qp_penalties  = C ;    %was 0.01
   C_qp_smoothness = C ;    %was 0.01
   C_lp_smoothness = 0.00 ; %linear problem parameters
   C_qp_smallness  = 1e-6 ; %was 1e-6 
   column_eps      = 1e-3 ;

   Q=zeros(2*126+N) ; % 1/2 * res' * Q * res
   Q(1:90,1:90)           = C_qp_smallness*diag(ones(1,90)) ;
   Q(91:126,91:126)       = C_qp_penalties*diag(ones(1,36)) ;
   Q(127:2*126,127:2*126) = C_qp_smoothness*diag(ones(1,126)) ;
   f = [zeros(1,126) C_lp_smoothness*ones(1,126) ones(1,N)/N] ; % <res, f'>

   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   % qp_solve: LB <= res <= UB
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   INF = 1e20 ;
   LB = [-INF*ones(1,126) zeros(1,126) zeros(1,N)] ; % lower bound for res
   UB = [ INF*ones(1,2*126) INF*ones(1,N)] ; % upper bound for res
   """
