# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008 Fabio De Bona
# Copyright (C) 2008 Max-Planck-Society

import sys
import pdb
import math
import os.path

import pymosek 
import mosekarr 

# Define a stream printer to grab output from MOSEK 
#def streamprinter(text): print text,

jp = os.path.join


def createTaskObject(stream_func):
   mskhandle = pymosek.mosek()
   env = mskhandle.Env()
   env.init () 
   task = env.Task(0,0) 
   task.set_Stream(pymosek.streamtype.log, stream_func)
   return task


def enforceMonotonicityBetween(begin,end):
   """
   This function returns constraints that 
 
   x_i <= x_i+1 for i in range begin,end

   Note that we use the row-based description of mosek
   """

   num_rows = end - begin

   # the entries of the constraint term, here
   # x_i - x_i+1 for all num_rows rows
   aval = mosekarr.array([1.0,-1.0]*num_rows)

   indices = [0]*(num_rows*2)
   ctr = 0
   for idx in range(begin,end):
      indices[ctr] = idx
      indices[ctr+1] = idx+1
      ctr += 2

   # column indices of each aval entry
   asub = mosekarr.array(indices)

   aptrb = mosekarr.array(range(0,num_rows*2,2))
   aptre = mosekarr.array(range(2,num_rows*2+1,2))

   bkc = mosekarr.array([pymosek.boundkey.up]*num_rows)
   blc = mosekarr.array([float("-inf")]*num_rows)
   buc = mosekarr.array([0.0]*num_rows)

   return aptrb,aptre,asub,aval,bkc,blc,buc


def createSmoothnessRegularizer(numFeatures,settings):
   """
   Creates the quadratic term of our objective function.
   """

   P = mosekarr.array([0.0]*numFeatures**2)
   # for numerical stability
   for i in range(numFeatures):
      P[i*numFeatures+i] = 1e-5

   lengthSP    = settings['numLengthSuppPoints']
   donSP       = settings['numDonSuppPoints']
   accSP       = settings['numAccSuppPoints']
   mmatrixSP   = settings['matchmatrixRows']*settings['matchmatrixCols']
   numq        = settings['numQualSuppPoints']
   numQPlifs   = settings['numQualPlifs'] 

   intronLengthGroupParam  = 1.0
   donorGroupParam         = 1.0
   acceptorGroupParam      = 1.0
   matchmatrixGroupParam   = 1.0
   qualityGroupParam       = 1.0

   # add off-diagonal terms for intron length parameters
   for j in range(lengthSP-1):
      P[j*numFeatures+j+1]       += -intronLengthGroupParam
      P[(j+1)*numFeatures+j]     += -intronLengthGroupParam
      P[j*numFeatures+j]         +=  intronLengthGroupParam
      P[(j+1)*numFeatures+j+1]   +=  intronLengthGroupParam

   # add off-diagonal terms for donor parameters
   for j in range(lengthSP,lengthSP+donSP-1):
      P[j*numFeatures+j+1]       += -donorGroupParam
      P[(j+1)*numFeatures+j]     += -donorGroupParam
      P[j*numFeatures+j]         +=  donorGroupParam
      P[(j+1)*numFeatures+j+1]   +=  donorGroupParam
   
   # add off-diagonal terms for acceptor parameters
   for j in range(lengthSP+donSP,lengthSP+donSP+accSP-1):
      P[j*numFeatures+j+1]       += -acceptorGroupParam
      P[(j+1)*numFeatures+j]     += -acceptorGroupParam
      P[j*numFeatures+j]         +=  acceptorGroupParam
      P[(j+1)*numFeatures+j+1]   +=  acceptorGroupParam

   # no off-diagonal terms for matchmatrix parameters
   for j in range(lengthSP+donSP+accSP,lengthSP+donSP+accSP+mmatrixSP):
       P[j*numFeatures+j] += matchmatrixGroupParam

   # add off-diagonal terms for quality parameters
   offset = lengthSP+donSP+accSP+mmatrixSP
   for k in range(numQPlifs):
      begin = offset+(k*numq)
      end   = offset+((k+1)*numq)

      for j in range(begin,end-1):
         P[j*numFeatures+j+1]       += -qualityGroupParam
         P[(j+1)*numFeatures+j]     += -qualityGroupParam
         P[j*numFeatures+j]         +=  qualityGroupParam
         P[(j+1)*numFeatures+j+1]   +=  qualityGroupParam

   subi_list   = []
   subj_list   = []
   values      = [] 
   for i in range(numFeatures):
      for j in range(i,numFeatures):
         if i > j:
            continue

         if math.fabs(P[i*numFeatures+j]) > 0.0:
            # switch indices here as pymosek expects the _lower_ triangular part
            # of the symmetric quadratic matrix
            subi_list.append(j)
            subj_list.append(i)
            values.append(P[i*numFeatures+j])

   subi  = mosekarr.array(subi_list)
   subj  = mosekarr.array(subj_list)
   P_val = mosekarr.array(values)
   return subi,subj,P_val


def createL2Regularizer(numFeatures):
   # add the quadratic term of the objective function
   index = range(numFeatures)
   subi  = mosekarr.array(index)
   subj  = mosekarr.array(index)
   P_val = mosekarr.array([1.0]*numFeatures)
   return subi,subj,P_val


class SIQPSolver():
   """
   This class is a wrapper for the cvxopt qp solver.
   It has the purpose to support an iterative addition of
   constraints to the original problem.
   
   """

   def __init__(self,numFeatures,numExamples,working_dir,settings,epsilon=10e-4):
      assert numFeatures > 0, 'You have to have at least one feature!'
      assert numExamples > 0, 'You have to have at least one example!'
      self.numFeatures = numFeatures
      self.numExamples = numExamples
      self.C = settings['C']
      self.epsilon = epsilon
      self.working_dir = working_dir

      # first set the logfile as all messages have to be written somewhere
      log_fn = jp(self.working_dir,'mosek_status.log')
      try:
         self.log_fh = open(log_fn,'w+')
      except:
         print 'ERROR: Could not create solver logging file:\n %s!' % log_fn

      self.plog("Initializing SIQPSolver...")
      self.plog("Feature space dim. is %d. Number of examples: %d" % (self.numFeatures,self.numExamples))
      self.plog("Regularization parameters are C=%f."%(self.C))

      self.old_objective_value = -(sys.maxint-1)

      # specify the number of variables and number of constraints of our
      # initial problem
      self.numvar = self.numFeatures + self.numExamples
      self.numcon = 0

      try:
         self.task = createTaskObject(self.plog)
      except:
         self.plog('ERROR: Could not fetch MOSEK environment and create task object!')
   
      # create the linear term of the objective
      self.q = mosekarr.array([0.0]*self.numFeatures + [1.0]*self.numExamples)
      self.q = self.C * self.q

      # Bound keys for variables
      # -> add bounds to enforce non-negativity of slacks
      # bound keys for variables 
      # the parameter file consists of w and slacks 
      # the parameters w are unbounded whereas the slacks are bounded below by
      # 0 and unbounded above
      bkx = mosekarr.array([pymosek.boundkey.fr]*self.numFeatures + [pymosek.boundkey.lo]*self.numExamples)

      # lower bounds on variables
      blx = mosekarr.array([float("-inf")]*self.numFeatures + [0.0]*self.numExamples)

      # upper bounds on variables
      bux = mosekarr.array([float("+inf")]*(self.numFeatures+self.numExamples))
      
      # quadratic term creation should take place after above constraint system
      # generation because we also create monotonicity contraints in there
      
      # initialize a zero constraint matrix of 1xnumvars as mosek does not
      # accept initially empty constraint matrices
      asub = mosekarr.array([0]*self.numvar)
      aval = mosekarr.array([0.0]*self.numvar)
      ptrb = mosekarr.array(range(0,self.numvar))
      ptre = mosekarr.array(range(1,self.numvar+1))

      bkc = mosekarr.array([pymosek.boundkey.fr])
      blc = mosekarr.array([float("-inf")])
      buc = mosekarr.array([float("+inf")])

      self.task.inputdata(self.numcon,self.numvar,
         self.q,  # linear term
         0.0,     # objective fixed value
         ptrb,ptre,
         asub,aval,
         bkc,blc,buc,      # constraint bounds
         bkx,blx,bux,      # variable bounds
         )

      self.task.putobjsense(pymosek.objsense.minimize)
   
      # create quadratic term
      subi,subj,P_val = createSmoothnessRegularizer(numFeatures,settings)
      #subi,subj,P_val = createL2Regularizer(numFeatures)

      # add quadratic term to objective
      self.task.putqobj(subi,subj,P_val)

      lengthSP    = settings['numLengthSuppPoints']
      donSP       = settings['numDonSuppPoints']
      accSP       = settings['numAccSuppPoints']

      #pdb.set_trace()
      # create constraints that enforce monotonic increasing donor scores
      aptrb,aptre,asub,aval,bkc,blc,buc = enforceMonotonicityBetween(lengthSP,lengthSP+donSP)

      # add donor constraints to the problem
      self.task.appendcons(aptrb,aptre,asub,aval,bkc,blc,buc)

      # create constraints that enforce monotonic increasing acceptor scores
      aptrb,aptre,asub,aval,bkc,blc,buc = enforceMonotonicityBetween(lengthSP+donSP,lengthSP+donSP+accSP)

      # add acceptor constraints to the problem
      self.task.appendcons(aptrb,aptre,asub,aval,bkc,blc,buc)


   def plog(self,string):
      if self.log_fh != None:
         self.log_fh.write(string)
         self.log_fh.flush()
      else:
         print string


   def addConstraint(self, phi_diff, idx):
      """
         phi_diff is defined as phi(x_i,y_i) - phi(x_i,yhat)
      """

      assert len(phi_diff) == self.numFeatures

      phi_diff = phi_diff.flatten().tolist()[0]

      loss = 1.0
      self.currentIdx = idx
      self.currentLoss = loss
      scalar_prod = 0.0

      # row-based description
      asub = mosekarr.array(range(self.numvar))
      slack_part = [0.0]*self.numExamples
      slack_part[idx] = 1.0
      aval = mosekarr.array(phi_diff+slack_part)

      aptrb = mosekarr.array([0])
      aptre = mosekarr.array([self.numvar])

      bkc = mosekarr.array([pymosek.boundkey.lo])
      blc = mosekarr.array([float(loss)])
      buc = mosekarr.array([float("+inf")])

      self.task.appendcons(aptrb,aptre,asub,aval,bkc,blc,buc)

      return True


   def solve(self):
      # for debugging purposes you can dump the problem to an ascii file 
      #self.task.writedata("problem_mosek.lp")
      #self.task.writedata("problem_mosek.mps")
      #self.task.writedata("problem_mosek.opf")
      self.task.optimize()
      self.task.solutionsummary(pymosek.streamtype.msg)

      self.prosta = []
      self.solsta = []
      solution_status = self.task.getsolutionstatus(pymosek.soltype.itr,self.prosta,self.solsta)

      primal_obj = self.task.getprimalobj(pymosek.soltype.itr)

      xx = mosekarr.zeros(self.numvar, mosekarr.Float)
      self.task.getsolutionslice(pymosek.soltype.itr,pymosek.solitem.xx,0,self.numvar,xx)

      num_dual_variables = self.task.getnumcon()
      yy = mosekarr.zeros(num_dual_variables, mosekarr.Float)
      self.task.getsolutionslice(pymosek.soltype.itr,pymosek.solitem.xc,0,num_dual_variables,yy)

      # make dual solution a member variable
      self.yy = yy

      return primal_obj,list(xx[0:self.numFeatures]),list(xx[self.numFeatures:])
