# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Written (W) 2008-2010 Fabio De Bona & Gunnar Raetsch
# Copyright (C) 2008-2010 Max-Planck-Society

import array
import cPickle
import os.path
import pdb
import sys
import traceback
import time

import numpy
from numpy.matlib import mat,zeros,ones,inf
from numpy.linalg import norm

try:
   import SIQP_CVXOPT
except:
   pass

try:
   import SIQP_MOSEK
except:
   pass


import QPalmaDP
from computeSpliceWeights import *
from set_param_palma import *
from computeSpliceAlignWithQuality import *
from TrainingParam import Param
from Plif import Plf,compute_donacc

from OutputFormat import createAlignmentOutput

from sequence_utils import SeqSpliceInfo,DataAccessWrapper,unbracket_seq
from utils import pprint_alignment, get_alignment

jp = os.path.join
dotProd = lambda x,y: (x.T * y)[0,0]


class SpliceSiteException:
   pass


def preprocessExample(training_set,exampleKey,seqInfo,settings):
   """ 
   This function...  
   """

   currentSeqInfo,originalRead,currentQualities,currentExons = training_set[exampleKey]
   id,chr,strand,up_cut,down_cut = currentSeqInfo

   read = unbracket_seq(originalRead)
   read = read.replace('-','')

   for c in ['A', 'C', 'G', 'T', 'N']:			 ###
	originalRead = originalRead.replace(c, c.lower()) ###

   ### dna, acc_supp, don_supp = seqInfo.get_seq_and_scores(id, strand, up_cut, down_cut, False, settings['perform_checks'])
   dna, acc_supp, don_supp = seqInfo.get_seq_and_scores(chr, strand, up_cut, down_cut, False, settings['perform_checks'])

   exons = currentExons - up_cut

   if exons.shape == (2,2):
      if exons[0,1]>exons[1,0]:
         print "swap exons:"
         print exons
         a=exons[0,0] ; exons[0,0]=exons[1,0] ; exons[1,0]=a ;
         a=exons[0,1] ; exons[0,1]=exons[1,1] ; exons[1,1]=a ;
         print exons
      fetched_dna_subseq = dna[exons[0,0]:exons[0,1]] + dna[exons[1,0]:exons[1,1]]
     
      donor_elem = dna[exons[0,1]:exons[0,1]+2]
      acceptor_elem = dna[exons[1,0]-2:exons[1,0]]

      if not ( donor_elem == 'gt' or donor_elem == 'gc' ):
         print 'preprocessExample: invalid donor in example %s'% exampleKey
         print donor_elem
         #pdb.set_trace()
         #raise SpliceSiteException

      if not ( acceptor_elem == 'ag' ):
         print 'preprocessExample: invalid acceptor in example %s'% exampleKey
         print acceptor_elem
         #raise SpliceSiteException

      #this assertion may not be true ... if there are indels
      #assert len(fetched_dna_subseq) == len(read), pdb.set_trace()

   return dna,read,acc_supp,don_supp,exons,originalRead,currentQualities


def performAlignment(dna,read,quality,mmatrix,donor,acceptor,ps,qualityPlifs,current_num_path,prediction_mode,settings):
   """
   Given the needed input this method calls the QPalma C module which
   calculates a dynamic programming in order to obtain an alignment
   """

   prb = QPalmaDP.createDoubleArrayFromList(quality)
   chastity = QPalmaDP.createDoubleArrayFromList([.0]*len(read))

   matchmatrix = QPalmaDP.createDoubleArrayFromList(mmatrix.flatten().tolist()[0])
   mm_len = settings['matchmatrixRows']*settings['matchmatrixCols']

   don_len = len(donor)
   donor = QPalmaDP.createDoubleArrayFromList(donor)
   acc_len = len(acceptor)
   acceptor = QPalmaDP.createDoubleArrayFromList(acceptor)

   # Create the alignment object representing the interface to the C/C++ code.
   currentAlignment = QPalmaDP.Alignment(settings['numQualPlifs'],settings['numQualSuppPoints'], settings['enable_quality_scores'])

   c_qualityPlifs = QPalmaDP.createPenaltyArrayFromList([elem.convert2SWIG() for elem in qualityPlifs])

   # calculates SpliceAlign, EstAlign, weightMatch, Gesamtscores, dnaest
   currentAlignment.myalign( current_num_path, dna, len(dna),\
    read, len(read), prb, chastity, ps, matchmatrix, mm_len, donor, don_len,\
    acceptor, acc_len, c_qualityPlifs, settings['remove_duplicate_scores'],\
    settings['print_matrix'] )

   if True: #prediction_mode:
      # part that is only needed for prediction
      result_len = currentAlignment.getResultLength()
      dna_array,read_array = currentAlignment.getAlignmentArraysNew()
      #print dna_array
      #print read_array
   else:
      dna_array = None
      read_array = None

   newSpliceAlign, newEstAlign, newWeightMatch, newDPScores, newQualityPlifsFeatures = currentAlignment.getAlignmentResultsNew()
   
   return newSpliceAlign, newEstAlign, newWeightMatch, newDPScores,\
   newQualityPlifsFeatures, dna_array, read_array


DEBUG = 1
ERROR = 2


class QPalma:
   """
   This class wraps the training and prediction functions for 
   the alignment.
   """
   
   def __init__(self,seqInfo,dmode=False):
      self.ARGS = Param()
      self.qpalma_debug_mode = dmode
      self.seqInfo = seqInfo
      self.logfh = None
      self.working_dir = None 
      self.status_prefix = ['','DEBUG: ','ERROR: ']


   def plog(self,string,status=0):
      if status == DEBUG and not self.qpalma_debug_mode:
         return 

      mes = '%s%s' % (self.status_prefix[status],string)

      if self.logfh != None:
         self.logfh.write(mes)
         self.logfh.flush()
      else:
         print mes


   def init_training(self,dataset_fn,result_dir,settings):
      self.working_dir = result_dir

      if not os.path.exists(self.working_dir):
         try:
            os.mkdir(self.working_dir)
         except:
            self.plog('Could not create working directory %s!'% self.working_dir,ERROR)
            sys.exit(1)

      # first set the logfile as all messages have to be written somewhere
      log_fn = jp(self.working_dir,'qpalma_train_status.log')
      try:
         self.logfh = open(log_fn,'w+')
      except:
         self.plog('Could not create QPalma logging file:\n %s!' % log_fn, ERROR)

      training_set = cPickle.load(open(dataset_fn))

      self.train(training_set,settings)


   def train(self,training_set,settings):
      """
      The mainloop for training.
      """

      numExamples = len(training_set)
      self.plog('Number of training examples: %d\n'% numExamples)
      print "debug" 

      self.noImprovementCtr = 0
      self.oldObjValue = 1e8
      objValue = 0
      objDiff = 1 

      remove_duplicate_scores = settings['remove_duplicate_scores']
      print_matrix            = settings['print_matrix']

      lengthSP    = settings['numLengthSuppPoints']
      donSP       = settings['numDonSuppPoints']
      accSP       = settings['numAccSuppPoints']
      mmatrixSP   = settings['matchmatrixRows']*settings['matchmatrixCols']
      numq        = settings['numQualSuppPoints']
      totalQualSP = settings['totalQualSuppPoints']

      # calculate the total number of features
      numFeatures = settings['numFeatures']

      # Initialize parameter vector
      param = numpy.matlib.rand(numFeatures,1)

      # we take the first quality vector of the tuple of quality vectors
      quality_index = 0

      # no intron length model
      if not settings['enable_intron_length']:
         param[:lengthSP] *= 0.0

      # Set the parameters such as limits penalties for the Plifs
      [h,d,a,mmatrix,qualityPlifs] = set_param_palma(param,settings)

      # Initialize solver 
      self.plog('Initializing problem...\n')
      
      try:
         if settings['solver_id'] == 0:
            solver = SIQP_CVXOPT.SIQPSolver(numFeatures,numExamples,self.working_dir,settings)
         elif settings['solver_id'] == 1:
            solver = SIQP_MOSEK.SIQPSolver(numFeatures,numExamples,self.working_dir,settings)
         #elif settings['solver_id'] == 3:
         #   pass
         else:
            assert False

      except:
         self.plog('Got no license. Telling queue to reschedule job...\n')
         #self.plog(str(sys.exc_info()))
         traceback.print_exc(file=self.logfh)
         sys.exit(99)

      #solver.enforceMonotonicity(lengthSP,lengthSP+donSP)
      #solver.enforceMonotonicity(lengthSP+donSP,lengthSP+donSP+accSP)

      # stores the number of alignments done for each example (best path, second-best path etc.)
      num_path = [settings['anzpath']]*numExamples

      currentPhi = zeros((numFeatures,1))
      totalQualityPenalties = zeros((totalQualSP,1))

      numConstPerRound = settings['numConstraintsPerRound']
      solver_call_ctr = 0

      iteration_nr = 0
      param_idx = 0
      const_added_ctr = 0

      featureVectors = zeros((numFeatures,numExamples))

      self.plog('Starting training...\n')
      # the main training loop
      while True:
         if iteration_nr >= settings['iter_steps']:
            break
         suboptimal_example = 0
         correct_spliced_alignments=0
         incorrect_spliced_alignments=0
         correct_unspliced_alignments=0
         incorrect_unspliced_alignments=0

         print 'Starting iteration %i (%d examples)' % (iteration_nr,len(training_set.keys()))
         for exampleIdx,example_key in enumerate(training_set.keys()):
###            if example_key % 1000 == 0:
###                print '   \rCurrent example %d ' % example_key,

            try:
               dna,read,acc_supp,don_supp,exons,original_read,currentQualities =\
               preprocessExample(training_set,example_key,self.seqInfo,settings)
            except SpliceSiteException:
               print "preprocessExample failed"
               continue

            if settings['enable_quality_scores']:
               quality = currentQualities[quality_index]
            else:
               quality = [40]*len(read)

            if not settings['enable_splice_scores']:
               for idx,elem in enumerate(don_supp):
                  if elem != -inf:
                     don_supp[idx] = 0.0

               for idx,elem in enumerate(acc_supp):
                  if elem != -inf:
                     acc_supp[idx] = 0.0

            # Compute feature vector of correct alignment
            trueSpliceAlign, trueWeightMatch, trueWeightQuality, dna_calc = computeSpliceAlign(dna, exons, read, original_read, quality, qualityPlifs,settings)

            dna_calc = dna_calc.replace('-','')

            #print 'right before computeSpliceWeights exampleIdx %d' % exampleIdx
            # Calculate the weights
            trueWeightDon, trueWeightAcc, trueWeightIntron = computeSpliceWeights(d, a, h, trueSpliceAlign, don_supp, acc_supp)
            if not settings['enable_intron_length']:
                for q in xrange(len(trueWeightIntron)):
                    trueWeightIntron[q]=0.0 ;
            trueWeight = numpy.vstack([trueWeightIntron, trueWeightDon, trueWeightAcc, trueWeightMatch, trueWeightQuality])

            currentPhi[0:lengthSP]                                            = mat(h.penalties[:]).reshape(lengthSP,1)
            if not settings['enable_intron_length']:
                for q in xrange(lengthSP):
                    assert currentPhi[q]==0.0
            currentPhi[lengthSP:lengthSP+donSP]                               = mat(d.penalties[:]).reshape(donSP,1)
            currentPhi[lengthSP+donSP:lengthSP+donSP+accSP]                   = mat(a.penalties[:]).reshape(accSP,1)
            currentPhi[lengthSP+donSP+accSP:lengthSP+donSP+accSP+mmatrixSP]   = mmatrix[:]

            if settings['enable_quality_scores']:
               totalQualityPenalties = param[-totalQualSP:]
               currentPhi[lengthSP+donSP+accSP+mmatrixSP:]                    = totalQualityPenalties[:]

            # Calculate w'phi(x,y) the total score of the alignment
            trueAlignmentScore = (trueWeight.T * currentPhi)[0,0]

            # The allWeights vector is supposed to store the weight parameter
            # of the true alignment as well as the weight parameters of the
            # num_path[exampleIdx] other alignments
            allWeights = zeros((numFeatures,num_path[exampleIdx]+1))
            allWeights[:,0] = trueWeight[:,0]

            AlignmentScores = [0.0]*(num_path[exampleIdx]+1)
            AlignmentScores[0] = trueAlignmentScore

            ################## Calculate wrong alignment(s) ######################
            # Compute donor, acceptor with penalty_lookup_new
            # returns two double lists
            donor, acceptor = compute_donacc(don_supp, acc_supp, d, a)

            ps = h.convert2SWIG()

            newSpliceAlign, newEstAlign, newWeightMatch, newDPScores, newQualityPlifsFeatures, unused1, unused2 =\
                            performAlignment(dna,read,quality,mmatrix,donor,acceptor,ps,qualityPlifs,num_path[exampleIdx],False,settings)

            current_prediction=self.calc_alignment(dna, read, quality, don_supp, acc_supp, d, a, h, mmatrix, qualityPlifs, settings)
            true_exons = []
            for x in exons:
                true_exons.append(x[0,0]) 
                true_exons.append(x[0,1])
            if not true_exons == current_prediction['predExons']:
                print "predicted alignment: ", current_prediction['DPScores'], current_prediction['predExons']
                print "true alignment: ", trueAlignmentScore, true_exons
                if len(true_exons)>2:
                    incorrect_spliced_alignments += 1
                else:
                    incorrect_unspliced_alignments += 1
            else:
                if len(true_exons)>2:
                    correct_spliced_alignments += 1
                else:
                    correct_unspliced_alignments += 1
            if exampleIdx%100 == 0:
                print "incorrect_spliced_alignments=%i/%i (%2.1f%%)" % (incorrect_spliced_alignments, incorrect_spliced_alignments+correct_spliced_alignments, 100*incorrect_spliced_alignments/(incorrect_spliced_alignments+correct_spliced_alignments+1e-6)),
                print "incorrect_unspliced_alignments=%i/%i (%2.1f%%)" % (incorrect_unspliced_alignments, incorrect_unspliced_alignments+correct_unspliced_alignments, 100*incorrect_unspliced_alignments/(incorrect_unspliced_alignments+correct_unspliced_alignments+1e-6))

            mm_len = settings['matchmatrixRows']*settings['matchmatrixCols']

            # check for correct reshaping !!
            newSpliceAlign = numpy.mat(newSpliceAlign).reshape(num_path[exampleIdx],len(dna))
            newWeightMatch = numpy.mat(newWeightMatch).reshape(num_path[exampleIdx],mm_len)

            newQualityPlifsFeatures = numpy.mat(newQualityPlifsFeatures).reshape(num_path[exampleIdx],settings['totalQualSuppPoints'])
            # Calculate weights of the respective alignments. Note that we are calculating n-best alignments without 
            # hamming loss, so we have to keep track which of the n-best alignments correspond to the true one in order 
            # not to incorporate a true alignment in the
            # constraints. To keep track of the true and false alignments we
            # define an array true_map with a boolean indicating the
            # equivalence to the true alignment for each decoded alignment.
            true_map = [0]*(num_path[exampleIdx]+1)
            true_map[0] = 1

            for pathNr in range(num_path[exampleIdx]):
               weightDon, weightAcc, weightIntron = computeSpliceWeights(d, a, h, newSpliceAlign[pathNr,:].flatten().tolist()[0], don_supp, acc_supp)
               if not settings['enable_intron_length']:
                   for q in xrange(len(weightIntron)):
                       weightIntron[q]=0.0 ;
               
               decodedQualityFeatures = zeros((settings['totalQualSuppPoints'],1))
               decodedQualityFeatures = newQualityPlifsFeatures[pathNr,:].T
               # Gewichte in restliche Zeilen der Matrix speichern
               allWeights[:,pathNr+1] = numpy.vstack([weightIntron, weightDon, weightAcc, newWeightMatch[pathNr,:].T, decodedQualityFeatures[:]])

               hpen = mat(h.penalties).reshape(len(h.penalties),1)
               dpen = mat(d.penalties).reshape(len(d.penalties),1)
               apen = mat(a.penalties).reshape(len(a.penalties),1)
               features = numpy.vstack([hpen, dpen, apen, mmatrix[:], totalQualityPenalties[:]])

               featureVectors[:,exampleIdx] = allWeights[:,pathNr+1]

               AlignmentScores[pathNr+1] = (allWeights[:,pathNr+1].T * features)[0,0]

               distinct_scores = False
               if math.fabs(AlignmentScores[pathNr] - AlignmentScores[pathNr+1]) > 1e-5:
                  distinct_scores = True

               # Check wether scalar product + loss equals viterbi score
               if not math.fabs(newDPScores[pathNr] - AlignmentScores[pathNr+1]) <= 1e-5:
                  self.plog("Scalar prod. + loss does not equal Viterbi score!\n")
                  self.plog(" scalar prod (correct) : %f\n"%AlignmentScores[0])
                  self.plog(" scalar prod (pred.) : %f %f\n"%(newDPScores[pathNr],AlignmentScores[pathNr+1]))
                  pdb.set_trace()

               # if the pathNr-best alignment is very close to the true alignment consider it as true
               if norm( allWeights[:,0] - allWeights[:,pathNr+1] ) < 1e-5:
                  true_map[pathNr+1] = 1

               if not trueAlignmentScore <= max(AlignmentScores[1:]) + 1e-6:
                  print "suboptimal_example %d (%s)\n" % (exampleIdx,example_key)
                  #trueSpliceAlign, trueWeightMatch, trueWeightQuality
                  #dna_calc=computeSpliceAlign(dna, exons, est, trueSpliceAlignoriginal_est, quality, qualityPlifs)

                  print exons
                  suboptimal_example += 1
                  self.plog("suboptimal_example %d (%s)\n" % (exampleIdx,example_key))
                  #pdb.set_trace()

               # the true label sequence should not have a larger score than the maximal one 
               # this means that all n-best paths are to close to each other 
               # we have to extend the n-best search to a (n+1)-best
               if len([elem for elem in true_map if elem == 1]) == len(true_map):
                  num_path[exampleIdx] = num_path[exampleIdx]+1

            # Choose true and first false alignment for extending
            firstFalseIdx = -1
            for map_idx,elem in enumerate(true_map):
                if elem == 0:
                    firstFalseIdx = map_idx
                    break
                
            if False:
                self.plog("Is considered as: %d\n" % true_map[1])
                dna_array,est_array = currentAlignment.getAlignmentArraysNew()
                _newSpliceAlign = newSpliceAlign[0].flatten().tolist()[0]
                _newEstAlign = newEstAlign[0].flatten().tolist()[0]
                
            # if there is at least one useful false alignment add the
            # corresponding constraints to the optimization problem
            if firstFalseIdx != -1:
                firstFalseWeights = allWeights[:,firstFalseIdx]
                differenceVector  = trueWeight - firstFalseWeights
                const_added = solver.addConstraint(differenceVector, exampleIdx)
                const_added_ctr += const_added

               # end of one example processing 

            # call solver every nth example / added constraint
            if (const_added_ctr+solver_call_ctr+1) % numConstPerRound == 0 or exampleIdx == len(training_set.keys())-1:
               t0=time.clock() 
               objValue,w,self.slacks = solver.solve()
               print time.clock() - t0, "seconds solver time"
               # increase numConstPerRound, if total solve time is larger than 100 seconds ... #hack
               if (time.clock() - t0)*len(training_set.keys())/numConstPerRound > 100: 
                   new_numConstPerRound = numConstPerRound*2
                   if new_numConstPerRound > len(training_set.keys()):
                       new_numConstPerRound = 1000000 # this (essentially) disables the mod-based invokation of solver
                   if new_numConstPerRound != numConstPerRound:
                       print "increasing numConstPerRound from %i to %i" % (numConstPerRound, new_numConstPerRound)
                       numConstPerRound = new_numConstPerRound ;
               solver_call_ctr += 1

               objDiff = math.fabs(objValue - self.oldObjValue)
               self.oldObjValue = objValue
               print "objValue is %f" % objValue
               
               sum_xis = 0
               for elem in self.slacks:
                   sum_xis +=  elem
                   
               print 'sum of slacks is %f'% sum_xis 
               self.plog('sum of slacks is %f\n'% sum_xis)
                   
               for i in range(len(param)):
                   param[i] = w[i]
               if not settings['enable_intron_length']:
                   for q in xrange(lengthSP):
                       assert math.fabs(param[q])<1e-5

               fd=open(jp(self.working_dir,'param_%d.pickle'%param_idx),'w+')
               cPickle.dump(param, fd)
               fd.close()
               
               param_idx += 1
               [h,d,a,mmatrix,qualityPlifs] =set_param_palma(param, settings)
               

            if objDiff <= 1e-6:
                self.noImprovementCtr += 1
            else:
                self.noImprovementCtr = 1 ;

            if self.noImprovementCtr > 2*numExamples:
                break

         # end of one iteration through all examples
         self.plog("suboptimal rounds %d\n" % suboptimal_example)
         print("suboptimal rounds %d\n" % suboptimal_example)

         if self.noImprovementCtr > numExamples*2:
             break

         iteration_nr += 1

      #
      # end of optimization 
      #  
      fd=open(jp(self.working_dir,'param_final.pickle'), 'w+')
      cPickle.dump(param, fd)
      fd.close()
      result={'h':h, 'd':d, 'a':a, 'mmatrix':mmatrix, 'qualityPlifs':qualityPlifs, 'settings':settings}
      self.FinalizeTraining(result, self.working_dir+'/parameters_final', iteration_nr, settings)

   def write_plif(self, fd, name, plif):
      fd.write('%s:\t%i\t%i\t%s\t' % (name, plif.min_len, plif.max_len, plif.transform))
      for i in xrange(len(plif.limits)):
          fd.write('%f,' % plif.limits[i])
      fd.write('\t') ;
      for i in xrange(len(plif.penalties)):
          fd.write('%f,' % plif.penalties[i])
      fd.write('\n') ;

   def write_matrix(self, fd, name, matrix):
      fd.write('%s:\t%i\t%i\t' % (name, matrix.shape[0], matrix.shape[1]))
      for i in xrange(matrix.shape[0]):
          for j in xrange(matrix.shape[1]):
              fd.write('%f,' % matrix[i,j])
      fd.write('\n') ;

   def write_scalar(self, fd, name, matrix):
      fd.write('%s:\t%i\t%i\t' % (name, 1, 1))
      fd.write('%f,' % matrix)
      fd.write('\n') ;

   def FinalizeTraining(self,param,name,iteration_nr, settings):
      self.plog("Training completed after %i iterations\nSaving final parameters to %s.{pickle,qpalma}" % (iteration_nr,name))
      print("Training completed after %i iterations\nSaving final parameters to %s.{pickle,qpalma}" % (iteration_nr,name))
      fd = open(name+'.pickle','w+')
      cPickle.dump(param,fd)
      fd.close()

      fd = open(name+'.qpalma', 'w+')
      self.write_plif(fd, 'h', param['h']) ;
      self.write_plif(fd, 'd', param['d']) ;
      self.write_plif(fd, 'a', param['a']) ;
      for i in xrange(len(param['qualityPlifs'])):
          self.write_plif(fd, 'q[%i]' % i, param['qualityPlifs'][i]) ;
      self.write_matrix(fd, 'mmatrix', param['mmatrix'])
      self.write_scalar(fd, 'prb_offset', settings['prb_offset'])
      fd.close() 
      
      if self.logfh != None:
         self.logfh.close()
         self.logfh = None 
   

###############################################################################
#
# End of the code needed for training 
#
# Begin of code for prediction
#
###############################################################################


   def init_prediction(self,dataset_fn,param_fn,result_dir,settings):
      """
      Performing a prediction takes...
      """

      if not os.path.exists(result_dir):
         try:
            os.mkdir(result_dir)
         except:
            self.plog('Could not create working directory %s!'% self.working_dir,ERROR)
            sys.exit(1)

      # first set the logfile as all messages have to be written somewhere
      log_fn = jp(result_dir,'qpalma_align_status.log')
      try:
         self.logfh = open(log_fn,'w+')
      except:
         self.plog('Could not create QPalma logging file:\n %s!' % log_fn, ERROR)

      # load dataset and fetch instances that shall be predicted
      dataset = cPickle.load(open(dataset_fn))
      print 'param_fn'
      print param_fn
      
      self.param = cPickle.load(open(param_fn))
      self.predict(dataset,result_dir,settings)

   def predict(self,prediction_set,result_dir,settings):
      """
      This method...
      """

      # Load parameter vector to predict with
      param = self.param 

      # Set the parameters such as limits/penalties for the Plifs
      [h,d,a,mmatrix,qualityPlifs] =\
      set_param_palma(param, settings)

      self.plog('Starting prediction...\n',DEBUG)

      self.problem_ctr = 0

      # where we store the predictions
      allPredictions = []
      cnt=0 

      # beginning of the prediction loop
      for example_key in prediction_set.keys():
         cnt=cnt+1 ;
         if cnt % 1000 ==0:
             print '   \rCurrent example %d (%s) ' % (cnt, example_key)
         for example in prediction_set[example_key]:

            currentSeqInfo,read,currentQualities = example

            id,chromo,strand,genomicSeq_start,genomicSeq_stop =\
            currentSeqInfo 

            if read.find('[') != -1:
               read = unbracket_seq(read)

            self.plog('Loading example id: %s...\n'% id,DEBUG)

            if settings['enable_quality_scores']:
               quality = currentQualities
            else:
               quality = [40]*len(read)

            if True:#try:
               currentDNASeq, currentAcc, currentDon = self.seqInfo.get_seq_and_scores(chromo, strand, genomicSeq_start, \
                                                                                       genomicSeq_stop, False, settings['perform_checks'])
            else:#except:
               pdb.set_trace()
               self.problem_ctr += 1
               print sys.exc_info()
               continue

            if not settings['enable_splice_scores']:
               for idx,elem in enumerate(currentDon):
                  if elem != -inf:
                     currentDon[idx] = 0.0

               for idx,elem in enumerate(currentAcc):
                  if elem != -inf:
                     currentAcc[idx] = 0.0

            current_prediction = self.calc_alignment(currentDNASeq, read, quality, currentDon, currentAcc, d, a, h, mmatrix, qualityPlifs,settings)

            current_prediction['id']         = id
            current_prediction['quality']    = ''.join(['%c' % (q+settings['prb_offset']) for q in quality])
            current_prediction['chr']        = chromo
            current_prediction['strand']     = strand
            current_prediction['start_pos']  = genomicSeq_start

            allPredictions.append(current_prediction)


      if not self.qpalma_debug_mode:
         self.FinalizePrediction(allPredictions,result_dir)
      else:
         return allPredictions

      createAlignmentOutput(settings,jp(result_dir,'all_predictions.pickle'),jp(result_dir,'all_predictions.blat'))

   def FinalizePrediction(self,allPredictions,result_dir):
      """ End of the prediction loop we save all predictions in a pickle file and exit """

      cPickle.dump(allPredictions,open(jp(result_dir,'all_predictions.pickle'),'w+'))
      self.plog('Prediction completed\n')
      mes =  'Problem ctr %d' % self.problem_ctr
      self.plog(mes+'\n')
      self.logfh.close()


   def calc_alignment(self, dna, read, quality, don_supp, acc_supp, d, a, h, mmatrix, qualityPlifs,settings):
      """
      Given two sequences and the parameters we calculate on alignment
      """

      donor, acceptor = compute_donacc(don_supp, acc_supp, d, a)

      if '-' in read:
         self.plog('found gap\n')
         read = read.replace('-','')
         #assert len(read) == Conf.read_size

      ps = h.convert2SWIG()

      newSpliceAlign, newEstAlign, newWeightMatch, newDPScores,\
      newQualityPlifsFeatures, dna_array, read_array =\
      performAlignment(dna,read,quality,mmatrix,donor,acceptor,ps,qualityPlifs,1,True,settings)

      mm_len = settings['matchmatrixRows']*settings['matchmatrixCols']

      true_map    = [0]*2
      true_map[0] = 1
      pathNr      = 0

      _newSpliceAlign   = array.array('B',newSpliceAlign)
      _newEstAlign      = array.array('B',newEstAlign)
       
      #(qStart, qEnd, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds, tExonSizes, tStarts, tEnds)
      alignment = get_alignment(_newSpliceAlign,_newEstAlign, dna_array, read_array) 

      dna_array   = array.array('B',dna_array)
      read_array  = array.array('B',read_array)

      newExons = self.calculatePredictedExons(newSpliceAlign)

      current_prediction = {'predExons':newExons, 'dna':dna, 'read':read, 'DPScores':newDPScores,\
      'alignment':alignment,'spliceAlign':_newSpliceAlign,'estAlign':_newEstAlign,\
      'dna_array':dna_array, 'read_array':read_array }

      return current_prediction


   def calculatePredictedExons(self,SpliceAlign):
      newExons = []
      oldElem = -1
      SpliceAlign.append(-1)
      for pos,elem in enumerate(SpliceAlign):
         if pos == 0:
            oldElem = -1
         else:
            oldElem = SpliceAlign[pos-1]

         if oldElem != 0 and elem == 0: # start of exon
            newExons.append(pos)

         if oldElem == 0 and elem != 0: # end of exon
            newExons.append(pos)

      return newExons
