#!/bin/bash

PROG=`basename $0`
DIR=`dirname $0`


. ${DIR}/../../bin/qpalma_config.sh

. $QPALMA_BIN_PATH/qpalma_header.sh $0 $*

set -e 

export PYTHONPATH=$PYTHONPATH:$QPALMA_TOOLS_PATH/errormodel:$QPALMA_TOOLS_PATH/gff:$QPALMA_TOOLS_PATH/sequences

if [ "$#" == "0" ];
then
	echo usage $0 \<output_dir\> \<error_model\> \<genome_fasta\> \<annotation_gff\> \<annotation_source_string\> \<read_length\> \<num_of_transcripts\> \<number_of_reads\> [\<paired\>]
	exit -1
fi

### INPUT ###
outdir=$1
errormodel=$2
genome=$3
annotation=$4
annotation_source=$5

### READ PARAMETERS ###
readlength=$6
transcripts=$7
samplesize=$8

if [ "$9" == "paired" ]
    then paired=" -p "
    else paired=""
fi

qualitysample=${10}

if [ ! -f $errormodel ];
then
	echo error model file $errormodel does not exist
	exit -1
fi

if [ ! -f $qualitysample ];
then
	echo quality sample file $qualitysample does not exist
	exit -1
fi

echo \# outdir=$outdir
echo \# errormodel=$errormodel
echo \# genome=$genome
echo \# annotation=$annotation
echo \# readlength=$readlength
echo \# transcripts=$transcripts
echo \# samplesize=$samplesize

### OUTPUT ###

if [ "$ARTREADS_optimal" != "" ]
    then optimal=$ARTREADS_optimal
    else optimal=$outdir/optimal_alignment.bed
fi
if [ "$ARTREADS_reads" != "" ]
    then output=$ARTREADS_reads
    else output=$outdir/art_reads.fastq
fi
if [ "$ARTREADS_trainingset" != "" ]
    then trainingset=$ARTREADS_trainingset
    else trainingset=$outdir/qpalma_train.bed
fi

if [ "$outdir" == "-" ]
then
    outdir=`dirname $optimal`
fi

mkdir -p $outdir

if [ "$ARTREADS_del_tmp" != "" ]
    then delete_tmp_output=$ARTREADS_del_tmp
    else delete_tmp_output=0
fi

if [ "$ARTREADS_noise" != "" ]
    then noise=$ARTREADS_noise
    else noise="0.0"
fi

if [ $delete_tmp_output -gt 0 ]
then
    tmpoutdir=`mktemp -d` || exit 1
else
    tmpoutdir=$outdir
fi

### INTERMEDIATE OUTPUT ###
logfile=$tmpoutdir/read_sampling.log
readfile=$tmpoutdir/readfile.fa
genomeconfig=$tmpoutdir/genome.config
spliced=$tmpoutdir/art_reads_align.spliced
mapper_output=$tmpoutdir/art_reads_align.result
triggerset=$tmpoutdir/triggered_reads


### Create quality sample from original read files
#echo "creating quality sample from files in $fastq ..."
#fastq_tmpname=`tempfile`
#pdirnamedirname $0`/tools/subsample_fastq.py $fastq $qualitysamplesize $fastq_tmpname 
#cat $fastq_tmpname | grep "@" -B 1 | grep -v -e '^--' -e "^@" -e "^+"  > $qualitysample 
#rm $fastq_tmpname 
#echo "... done"
#echo ""

### Create Genome Information Object from sequence of original genome

if [ ! -d "$tmpoutdir/genome" ];
then
	echo "creating genome information object from $genome in $tmpoutdir ..."
	python $QPALMA_TOOLS_PATH/sequences/make_gio.py $genome $tmpoutdir 
	echo "... done"
	echo ""
fi

### Sample reads from genome.config object according to transcript annotation in *.gff
echo python $QPALMA_TOOLS_PATH/gff/parse_transcript_from_gff.py $paired -a $annotation --source=$annotation_source -g $genomeconfig -f $readfile -l $logfile -r $readlength -t $transcripts -s $samplesize 
python $QPALMA_TOOLS_PATH/gff/parse_transcript_from_gff.py $paired -a $annotation --source="$annotation_source" -g $genomeconfig -f $readfile -l $logfile -r $readlength -t $transcripts -s $samplesize 

### Create artificial short reads reads from readfile *.fa and quality file *.quality using error model *.pickle
## report created errors in fastq-file -> set option '-e'

if [ "$9" == "paired" ]
then
  echo python $QPALMA_TOOLS_PATH/errormodel/sample.py -p -q $qualitysample -f ${readfile}_1,${readfile}_2 -m $errormodel -l ${logfile}_1,${logfile}_2 -o $optimal \> $output 
  python $QPALMA_TOOLS_PATH/errormodel/sample.py -p -q $qualitysample -f ${readfile}_1,${readfile}_2 -m $errormodel -l ${logfile}_1,${logfile}_2 -o $optimal > $output 
else
  echo python $QPALMA_TOOLS_PATH/errormodel/sample.py -q $qualitysample -f $readfile -m $errormodel -l $logfile -o $optimal \> $output 
  python $QPALMA_TOOLS_PATH/errormodel/sample.py -q $qualitysample -n $noise -f $readfile -m $errormodel -l $logfile -o $optimal > $output
fi

### Align artificial reads to the genome and log those reads triggering QPALMA as training set

echo " "
echo "Building training set ..."
### Build up palmapper index, if neccessary
if [ ! -f $genome.cid ]
then
    if [ $delete_tmp_output -gt 0 ]
    then
        gmapper_index_dir=$tmpoutdir
        cp $genome $tmpoutdir/
    else
        gmapper_index_dir=`dirname $genome`
    fi
    ### create indices
    ${PALMAPPER_PATH}pmindex -v -i $gmapper_index_dir/`basename $genome`
else
	gmapper_index_dir=`dirname $genome`
fi

#export LD_LIBRARY_PATH=~/svn/projects/QPalma/dyn_prog/cpplib:~/software/shogun/lib:$LD_LIBRARY_PATH
echo ${PALMAPPER_PATH}palmapper -q $output -i $gmapper_index_dir/`basename $genome` -o $mapper_output -f bed -z 5 -log-triggered-reads $triggerset -seed-hit-cancel-threshold 10000 -index-precache
${PALMAPPER_PATH}palmapper -q $output -i $gmapper_index_dir/`basename $genome` -o $mapper_output -f bed -z 5 -log-triggered-reads $triggerset -seed-hit-cancel-threshold 10000 -index-precache
echo "...done"
echo " "

python `dirname $0`/genTrainingSet.py -a $optimal -o $trainingset -t $triggerset

if [ $delete_tmp_output -gt 0 ]
then
    rm -r $tmpoutdir
fi

#echo $outdir $optimal $trainingset 
echo
echo \# create_art_reads done
echo
