#!/usr/bin/python

"""This tool parses artificial reads from a given genome information object according to the 
    transcript annotation of a given gff-file. Read length, number of different transcript and sample 
    size can be defined, all together resulting in a specific coverage."""

import pdb
import genome_utils
from Bio.SeqFeature import SeqFeature

def coverage_preview(options):
    """Prints an overview table for choosing right settings to end up with a specific coverage. """    
    print '\nCoverage table for readlength %d and sample size %d\n' % (options.readlength, options.settings.samplesize)

    average_trans_size = [10, 20, 50, 100, 250, 500, 1000]
    no_of_transcripts = range(10, 90, 10)
    print 'av t-size | # -trans ->',
    for j in no_of_transcripts: 
        print '%5d' % j,
    print '\n',    
    print 80 * '-'

    for i in average_trans_size:
        print '%10d|%13s' % (100*i,''),
        for j in no_of_transcripts:
            single_coverage = (i * j * 100) / options.readlength
            print '%4d ' % (options.samplesize / single_coverage),
        print '\n',    

def parse_options(argv):

    """Parses options from the command line """

    import sys
    from optparse import OptionParser, OptionGroup

    parser = OptionParser()
    required = OptionGroup(parser, 'REQUIRED')
    required.add_option('-a', '--annotation', dest='gff_file', metavar='FILE', help='file containing the gene annotation in gff format', default='-')
    required.add_option('-f', '--fasta', dest='fasta', metavar='FILE', help='fasta file for storing the created reads', default='-')
    required.add_option('-g', '--genomeinfo', dest='gio_file', metavar='FILE', help='file containing the genome information objct (created with mGene)', default='-')
    required.add_option('-l', '--logfile', dest='logfile', metavar='FILE', help='file created while sampling reads to reads.fa with parse_reads_from_gff.py', default='-')
    optional = OptionGroup(parser, 'OPTIONAL')
    optional.add_option('-r', '--readlength', dest='readlength', metavar='INT', type='int', help='length of one short read', default='36')
    optional.add_option('-t', '--transcripts', dest='transcripts', metavar='INT', type='int', help='number of different transcripts to sample from [default: 10]', default = 10)
    optional.add_option('-s', '--samplesize', dest='samplesize', metavar='INT', type='int', help='number of reads to sample [ default: 10 000 ]', default=10000)
    optional.add_option('-c', '--coverage', dest='cov_only', action='store_true', help='only print coverage table for curr, readlen and diff. transcript lengths, do not sample [default: not set]', default=False)
    optional.add_option('-p', '--paired', dest='paired', action='store_true', help='create paired reads [default: not set]', default=False)
    optional.add_option('-i', '--insertsize', dest='insertsize', metavar='INT', type='int', help='insert size for paired reads [300], minimum is 20', default='300')
    optional.add_option('-R', '--random', dest='r_seed', metavar='INT', type='int', help='fixed seed for random operations, -1 for random system seed, [default: 23 ]', default=23)
    optional.add_option('-S', '--source', dest='source', metavar='STRING', help='defines source for annotation (second column in gff-file), in case of more than one source, provide a comma-separated list: SOURCE1,SOURCE2,SOURCE3,... , [ default: Coding_transcript ]', default='Coding_transcript')
    parser.add_option_group(required)
    parser.add_option_group(optional)

    (options, args) = parser.parse_args()
    
    if len(argv) < 5 or '-' in [options.gff_file, options.fasta, options.gio_file, options.logfile] or options.insertsize < 20:
        parser.print_help()
        sys.exit(2)

    return options

def filter_by_boundary(trans_dict, chrm, seq):
    """Filters given set of introns by checking if specified boundaries 
    in the given sequence according to the consensus of GT/C ... AG"""
   
    filtered = []
    u12_filtered = 0
    num_trans = 0 
    ### find all exons and check their boundaries
    for gene in trans_dict.features:
        num_trans += max(len(gene.sub_features), 1)
        for trans in gene.sub_features:
            contig = []
            for exon in trans.sub_features:
                if exon.type == 'exon':
                    contig.append((exon.location.nofuzzy_start, exon.location.nofuzzy_end))
            contig.sort(lambda u, v : u[0] - v[0])
           
            ### no reverse complements considered right now, since we only sample from forward strand
            for idx in range(len(contig) - 1):
                intron = seq[contig[idx][1]:contig[idx + 1][0]].upper()
                if (intron[:2] != 'GT' and intron[:2] != 'GC') or intron[-2:] != 'AG':
                    if (intron[:2] == 'AT' and intron[-2:] == 'AC' ):
                        u12_filtered += 1    
                    filtered.append((gene, trans))
                    break
    
    ### delete all transcripts with mal-formed exons
    num_removed=0
    for (gene, trans) in filtered:
        #print 'remove %s ' % str((gene, trans))
        gene.sub_features.remove(trans)
        num_removed=num_removed+1
        if len(gene.sub_features) == 0:
            trans_dict.features.remove(gene)

    print 'removed %d of %d transcripts (%i genes) in chromosome %s due to imperfect intron boundaries or incorrect GFF3 structures' % (len(filtered), num_trans, len(trans_dict.features), chrm)
    print '%d of the removed transcripts had U12-dependent AT-AC introns' % u12_filtered

    ### check, if we filtered everything from this chrm
    if len(trans_dict.features) == 0:
        return True
    return False


def fix_structure(trans_dict):
    """Adapts the structure of the trans_dict to different gff-file structures"""

    for idx in range(len(trans_dict.features)):
        if trans_dict.features[idx].type == 'exon':
            exon = trans_dict.features[idx]
            trans_dict.features[idx] = SeqFeature(exon.location, type = 'gene', strand = exon.strand, id = exon.id)
            trans_dict.features[idx].sub_features = [SeqFeature(exon.location, type = 'Transcript', strand = exon.strand, id = exon.id)] 
            trans_dict.features[idx].sub_features[0].sub_features = [exon]
        elif len(trans_dict.features[idx].sub_features) > 0 and trans_dict.features[idx].sub_features[0].type == 'exon':
            exon = trans_dict.features[idx]
            trans_dict.features[idx] = SeqFeature(exon.location, type = 'gene', strand = exon.strand, id = exon.id)
            trans_dict.features[idx].sub_features = [exon]
#
#    for idx in range(len(trans_dict.features)):
#        if trans_dict.features[idx].type == 'mRNA' or trans_dict.features[idx].type == 'Transcript':
#            trans = trans_dict.features[idx]
#            trans_dict.features[idx] = SeqFeature(trans.location, type = 'gene', strand = trans.strand, id = trans.id)
#            trans_dict.features[idx].sub_features = [trans] 

def filter_by_strand(trans_dict, chrm):
    """Filters given set of introns by keeping only those on the + strand."""

    filtered = []
    for trans in trans_dict[chrm].features:
        element = trans
        while element.type != 'exon' and len(element.sub_features) > 0:
            element = element.sub_features[0]
        #if trans.strand != 1 :
        if element.strand != 1 :
            filtered.append(trans)
    
    print '\nin Chromosome %s:\n%d genes are located on minus strand \nkeep only the remaining %d genes on plus strand' % (chrm, len(filtered), len(trans_dict[chrm].features) - len(filtered))
    for trans in filtered:
        trans_dict[chrm].features.remove(trans)
    
    ### check, if we filtered everything from this chrm
    if len(trans_dict[chrm].features) == 0:
        return True
    return False

def create_reads(gio, trans_dict, gene_to_sample, options, total_counter, fout, fout2, flog, flog2):
    """ Creates contigs from annotion in trans_dict and samples reads according to
    settings from the flatfiles specified in the gio-object"""

    import random
    import os
    
    if options.r_seed == -1:
        random.seed(os.urandom(128))
    else:
        random.seed(options.r_seed)
    
    if options.paired:
        ext_readlength = (2 * options.readlength) + options.insertsize
    else:
        ext_readlength = options.readlength + (options.readlength / 2)

    for chrm in trans_dict.keys():
        gi_file = open(gio.basedir + '/genome/' + chrm + '.flat', 'r')
        seq = gi_file.read()
        gi_file.close()
 
        fix_structure(trans_dict[chrm])
        
        ### Filter transcripts 
        ### keeping only those on + strand
        if filter_by_strand(trans_dict, chrm):
            continue # all transcripts were filtered out

        ### keeping only those having introns with  G[TC] .... AG
        if filter_by_boundary(trans_dict[chrm], chrm, seq):
            continue # all transcripts were filtered out
        
        print ""
	   
        ### adapt number of transcripts to sample to available number of transcripts
        _avail_trans = 0
        _correction = 0
        for _gene in trans_dict[chrm].features:
            _avail_trans += len(_gene.sub_features)
        if gene_to_sample[chrm] > _avail_trans:
            _correction = gene_to_sample[chrm] - (_avail_trans) 
        gene_to_sample[chrm] = _avail_trans
        curr_samplesize =  int(round(gene_to_sample[chrm] * ((options.samplesize / float(options.transcripts - _correction)) + 1) ))
        curr_reads_created = 0
        curr_gene_created = 0
        curr_trans_created = 0
        already_read = dict()

        while curr_trans_created < gene_to_sample[chrm]:
            #if len(trans_dict[chrm].features)==0: break ;
            gene_id = random.randrange(len(trans_dict[chrm].features))
           # while not already_read.has_key(gene_id):
            if already_read.has_key(gene_id):
                if len(already_read.keys()) < len(trans_dict[chrm].features):
                    continue
                else:
                    break
            else:
                already_read[gene_id] = 1
                gene = trans_dict[chrm].features[gene_id]
 
            total_counter['genes'] += 1
            curr_gene_created += 1
            for trans in gene.sub_features:
                if curr_trans_created >= gene_to_sample[chrm]:
                    continue
                curr_trans_created += 1
                total_counter['trans'] += 1
                cum_length = 0
                idx = 0
                contig = []
              
                ### keep only the exons
                for idx2 in xrange(len(trans.sub_features) - 1, -1, -1):
                    if trans.sub_features[idx2].type != 'exon':
                        trans.sub_features.remove(trans.sub_features[idx2])

                ### assemble exons to a contig
                for exon in trans.sub_features:
                    if exon.type == 'exon':
                        contig.append([cum_length, idx, exon.location.nofuzzy_start])
                        idx += 1
                
                contig.sort(lambda u, v : u[2] - v[2])
                
                ### determine cumulative length for sorted exons
                for exon in contig:
                    exon[0] = cum_length
                    cum_length += (trans.sub_features[exon[1]].location.nofuzzy_end - trans.sub_features[exon[1]].location.nofuzzy_start) 
                total_counter['trans_length'] += cum_length
                
                contig.reverse()

                if cum_length <= ext_readlength:
                    continue
                
                #print '\rsampling from %s' % trans.id
                ### sample reads from this transcript
                for i in xrange(curr_samplesize / gene_to_sample[chrm]):
                    tmp_start = random.randrange(0, cum_length - ext_readlength + 1)
                    last_exons = []
                    history = -1
                    for exon in contig:
                        if tmp_start >= exon[0]:
                            abs_start = [trans.sub_features[exon[1]].location.nofuzzy_start + tmp_start - exon[0]]
                            read = seq[abs_start[0] : min(abs_start[0] + ext_readlength, trans.sub_features[exon[1]].location.nofuzzy_end)]
                            used_exons = [exon[1]]
                            block_sizes = [len(read)]
                            while len(read) < ext_readlength:
                                abs_start.append(trans.sub_features[last_exons[history][1]].location.nofuzzy_start)
                                read += seq[abs_start[-history] : min(abs_start[-history] + ext_readlength - len(read), trans.sub_features[last_exons[history][1]].location.nofuzzy_end)]
                                used_exons.append(last_exons[history][1])
                                history += -1
                                #block_sizes.append(len(read) - block_sizes[-1])
                                block_sizes.append(len(read) - sum(block_sizes))
                            assert(sum(block_sizes) == len(read))

                            ### Modify Block sizes 
                            mod_block_sizes = []
                            block_sum = 0
                            idx = 0
                            next_block = 0
                            while block_sum < options.readlength:
                                block_sum += block_sizes[idx]
                                if options.readlength - block_sum > 0:
                                    mod_block_sizes.append(block_sizes[idx])
                                elif options.readlength - block_sum + block_sizes[idx] > 0:
                                    mod_block_sizes.append(options.readlength - block_sum + block_sizes[idx])
                                elif  options.readlength - block_sum + block_sizes[idx] == 0:
                                    next_block = abs_start[len(mod_block_sizes)]
                                idx += 1

                            if len(mod_block_sizes) > 1 : 
                                total_counter['spliced'] += 1

                            ### Write read_info to log-file
                            curr_reads_created += 1
                            total_counter['reads'] += 1
                            read_id = 'read_' + str(total_counter['reads'])
                            if options.paired:
                                read_id += '/1'
                            print >> flog, '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%d\t+' % (read_id, read[:options.readlength], chrm, trans.id, used_exons[:len(mod_block_sizes)], abs_start[:len(mod_block_sizes)], mod_block_sizes, read[options.readlength:options.readlength + 20], next_block)
                            break
                        last_exons.append(exon)
                    
                    if not options.paired:
                        print >> fout, '>read_%d' % total_counter['reads']
                    else:
                        print >> fout, '>read_%d/1' % total_counter['reads']
                    print >> fout, '%s' % read[:options.readlength]

                    ### handle paired read case
                    if options.paired:
                        ### Modify Block sizes - this time from the other direction 
                        mod_block_sizes = []
                        block_sum = 0
                        idx = len(block_sizes) - 1
                        next_block = 0
                        while block_sum < options.readlength:
                            block_sum += block_sizes[idx]
                            if options.readlength - block_sum > 0:
                                mod_block_sizes.append(block_sizes[idx])
                            elif options.readlength - block_sum + block_sizes[idx] > 0:
                                mod_block_sizes.append(options.readlength - block_sum + block_sizes[idx])
                            elif  options.readlength - block_sum + block_sizes[idx] == 0:
                                next_block = abs_start[-len(mod_block_sizes)]
                            idx -= 1
                        
                        ### adapt abs_start 
                        abs_start[-len(mod_block_sizes)] += (block_sizes[-len(mod_block_sizes)] - mod_block_sizes[0])

                        if len(mod_block_sizes) > 1 : 
                            total_counter['spliced'] += 1

                        ### Write read_info to log-file
                        read_id = 'read_' + str(total_counter['reads'])
                        if options.paired:
                            read_id += '/2'
                        _genome_start = abs_start[-len(mod_block_sizes)] #+ block_sizes[-len(mod_block_sizes)] - mod_block_sizes[-1]
                        print >> flog2, '%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%d\t-\t%s\t%s' % (read_id, genome_utils.reverse_complement(read)[:options.readlength], chrm, trans.id, used_exons[-len(mod_block_sizes):], abs_start[-len(mod_block_sizes):], mod_block_sizes, genome_utils.reverse_complement(read)[options.readlength:options.readlength + 20], next_block, options.insertsize, _genome_start)
                        print >> fout2, '>read_%d/2' % total_counter['reads']
                        print >> fout2, '%s' % genome_utils.reverse_complement(read)[:options.readlength]
        print '... created %d reads in %d transcripts / %d genes on chromosome %s\n' % (curr_reads_created, curr_trans_created, curr_gene_created, chrm)
        
def main():
    """This script parses a GFF file for a given number of transcripts. According to these annotations 
    different reads are sampled from the flat files specified by a genome.config object.
    """
    
    import sys
    import GFFParser
    import cPickle
    import random


    options = parse_options(sys.argv)
   
    if options.cov_only:
        coverage_preview(options)
        sys.exit(0)

    total_counter = dict()
    total_counter['reads'] = 0
    total_counter['trans'] = 0 
    total_counter['genes'] = 0
    total_counter['spliced'] = 0
    total_counter['trans_length'] = 0
    
    ### Read: genome.config
    gio = genome_utils.GenomeInfo(options.gio_file)

    ### Parse transcripts from Gff-File
    print 'parsing transcripts from GFF ...'
    iterator = GFFParser.GFFAddingIterator()
    examiner = GFFParser.GFFExaminer()
    id_dict = examiner.available_limits(options.gff_file)['gff_id']
    id_list = [key[0] for key in id_dict.keys()]
    ### Check, if genome.config fits the used gff-file
    try:
        id_list.sort()
        gio.contig_names.sort()
        assert(id_list == gio.contig_names)
    except AssertionError:
        print >> sys.stdout, \
        '\nWARNING: \n %s does not fit chromosome names in %s' % (options.gio_file, options.gff_file)
        skiplist = list(set(id_list).difference(set(gio.contig_names)))
        keeplist = list(set(id_list).intersection(set(gio.contig_names)))
        keeplist.sort()
        print >> sys.stdout, \
        'Will skip all missing or ambiguous chromosomes and keep only chromosomes %s !\n' % str(keeplist)
        for i in skiplist:
            id_list.remove(i)

    ### parse source strings from options
    source_strings = options.source.strip().split(',') 
    ### take all available sources, if no sources are given
    if source_strings == ['-']:
        source_dict = examiner.available_limits(options.gff_file)['gff_source_type']
        taken_sources = dict()
        for key in source_dict.keys():
            if key[1] in set(['gene', 'mRNA', 'Transcript', 'exon', 'intron']):
                if taken_sources.has_key(key[0]):
                    taken_sources[key[0]] += 1
                else:
                    taken_sources[key[0]] = 1
        if len(taken_sources.keys()) < 4:
            print >> sys.stderr, 'No suitable sources found!'
            sys.exit(-1)
        else:
            source_strings = taken_sources.keys()
    
    gff_sources = []
    #types = ['gene', 'mRNA', 'Transcript', 'exon', 'intron']
    ### only extract exon information and let the GFFParser infer the rest of the structure
    ### we can sanitize that later, but otherwise this leads to doubled structured in some cases
    types = ['exon']
    for source_string in source_strings:
        gff_sources.extend(zip([source_string] * len(types), types))

    if options.paired:    
        fout = open(options.fasta + '_1', 'w')
        fout2 = open(options.fasta + '_2', 'w')
    else:
        fout = open(options.fasta, 'w')
        fout2 = None
    if options.paired:
        flog = open(options.logfile + '_1', 'w')
        flog2 = open(options.logfile + '_2', 'w')
    else:
        flog = open(options.logfile, 'w')
        flog2 = None

    print 'sampling reads from transcripts ...'

    ### Build up sample list by scanning gff-file once
    sample_dict = dict()
    sample_list = []
    _target = set(['mRNA', 'Coding_transcript']) ### TODO CDS?
    for _line in open(options.gff_file):
        if _line[0] == '#':
            continue
        _sl = _line.strip().split('\t')
        if _sl[2] in _target:
            if sample_dict.has_key(_sl[0]):
                sample_dict[_sl[0]] += 1
            else:
                sample_dict[_sl[0]] = 1
    for key in sample_dict.keys():
        sample_list.extend(sample_dict[key] * [key])
    del sample_dict

    ### How many transcripts should be sampled per chromosome
    if len(sample_list) < options.transcripts:
        sample = []
        for i in range((options.transcripts / len(sample_list)) + 1):
            sample.extend(random.sample(sample_list, len(sample_list)))
    else:
        sample = random.sample(sample_list, options.transcripts)
   
    gene_to_sample = dict()
    for gid in id_list:
        gene_to_sample[gid] = sample.count(gid)

 #gff_handle = open(options.gff_file)

    ### create reads in each chromosome present in gene_to_sample
    id_counter = 0
    number = 0
    for gid in id_list:
        id_counter += 1
        if gene_to_sample[gid] == 0:
            print 'skip chromosome %s - 0 transcripts to sample' % gid
        else:
            trans_dict = dict()
            limits = {'gff_source_type':gff_sources, 'gff_id':[gid]}
            print '... parsing chromosome %s (%i / %i) ...' % (gid, id_counter, len(id_list))
            trans_dict[gid] = iterator.get_all_features(options.gff_file, limits)[gid]
            number += len(trans_dict[gid].features)
            
            ### Create contigs from transcript annotation and sample reads
            create_reads(gio, trans_dict, gene_to_sample, options, total_counter, fout, fout2, flog, flog2)
            
            del trans_dict
         
    ### print summary
    if options.paired:
        ext_readlength = (2 * options.readlength) + options.insertsize
    else:
        ext_readlength = options.readlength + (options.readlength / 2)
    print '...%d transcripts read\n' % number
    print '\n... in total: created %d reads in %d transcripts / %d genes' % (total_counter['reads'], total_counter['trans'], total_counter['genes'])
    print '... including %d spliced reads\n' % total_counter['spliced']
    print 'average transcript length (exonic segments only): %d' % (total_counter['trans_length'] / max(total_counter['trans'], 1))
    print 'approximate coverage: %4.2f' % (options.samplesize / max(1, float(total_counter['trans_length'])) * max(options.readlength, 1))
    print 'approximate coverage (ext): %4.2f' % (options.samplesize / max(1, float(total_counter['trans_length'])) * max(ext_readlength, 1))


#            for rec in GFF.parse(gff_handle, limits):
#                trans_dict[id] = rec
#        gff_handle.close()
 
    fout.close()
    if options.paired:
        fout2.close()
        flog2.close()
    flog.close()


if __name__ == '__main__':
    main()

